@@ -1,15 +1,18 @@
 
-import core
+import os
+import os.path
+import pprint
+
+from . import core
 import ctypes
 try:
     import cPickle as pickle
 except ImportError:
     import pickle
 
-import os
-import os.path
-
-
+import sys
+if 2==(sys.version_info[0]):
+  range=xrange
 RT_Memory = 0
 RT_Disk = 1
 RT_Custom = 2
@@ -24,7 +27,10 @@ RT_TPRTree = 2
 
 __c_api_version__ = core.rt.SIDX_Version()
 
-if (__c_api_version__.split('.')[1] < 7):
+major_version, minor_version, patch_version = [
+    int(t) for t in __c_api_version__.decode('utf-8').split('.')]
+
+if (major_version < 2 and minor_version < 7):
     raise Exception("This version of Rtree requires libspatialindex 1.7.0 or greater")
 
 __all__ = ['Rtree', 'Index', 'Property']
@@ -168,7 +174,7 @@ class Index(object):
         basename = None
         storage = None
         if args:
-            if isinstance(args[0], basestring):
+            if isinstance(args[0], str) or isinstance(args[0], bytes):
                 # they sent in a filename
                 basename = args[0]
                 # they sent in a filename, stream
@@ -185,10 +191,10 @@ class Index(object):
 
         if basename:
             self.properties.storage = RT_Disk
-            self.properties.filename = basename
+            self.properties.filename = str.encode(str(basename))
 
             # check we can read the file
-            f = basename + "." + self.properties.idx_extension
+            f = str.encode(str(basename) + "." + str(self.properties.idx_extension))
             p = os.path.abspath(f)
 
 
@@ -211,6 +217,9 @@ class Index(object):
                 message = "Unable to open file '%s' for index storage"%f
                 raise IOError(message)
         elif storage:
+            if (major_version < 2 and minor_version < 8):
+                raise core.RTreeError("libspatialindex {0} does not support custom storage".format(__c_api_version__))
+
             self.properties.storage = RT_Custom
             if storage.hasData:
                 self.properties.overwrite = bool(kwargs.get('overwrite', False))
@@ -254,6 +263,8 @@ class Index(object):
                     return
 
                 core.rt.Index_Destroy(self.handle)
+                self.owned = False
+                self.handle = None
 
     def close(self):
         """Force a flush of the index to storage. Renders index
@@ -264,7 +275,7 @@ class Index(object):
             self.handle = None
             self.owned = False
         else:
-            raise IOError, "Unclosable index"
+            raise IOError("Unclosable index")
 
     def get_coordinate_pointers(self, coordinates):
 
@@ -339,11 +350,11 @@ class Index(object):
 
         """
         p_mins, p_maxs = self.get_coordinate_pointers(coordinates)
+        data = ctypes.c_ubyte(0)
+        size = 0
+        pyserialized = None
         if obj is not None:
             size, data, pyserialized = self._serialize(obj)
-        else:
-            data = ctypes.c_ubyte(0)
-            size = 0
         core.rt.Index_InsertData(self.handle, id, p_mins, p_maxs, self.properties.dimension, data, size)
     add = insert
 
@@ -455,10 +466,10 @@ class Index(object):
 
         try:
             if objects != 'raw':
-                for i in xrange(num_results):
+                for i in range(num_results):
                     yield Item(self.loads, items[i])
             else:
-                for i in xrange(num_results):
+                for i in range(num_results):
                     data = _get_data(items[i])
                     if data is None:
                         yield data
@@ -476,7 +487,7 @@ class Index(object):
         its = ctypes.cast(items, ctypes.POINTER(ctypes.c_void_p))
 
         try:
-            for i in xrange(num_results):
+            for i in range(num_results):
                 yield items.contents[i]
             core.rt.Index_Free(its)
         except:
@@ -609,7 +620,7 @@ class Index(object):
 
         """
         assert len(interleaved) % 2 == 0, ("must be a pairwise list")
-        dimension = len(interleaved) / 2
+        dimension = len(interleaved) // 2
         di = []
         for i in range(dimension):
             di.extend([interleaved[i], interleaved[i + dimension]])
@@ -657,7 +668,7 @@ class Index(object):
             than 0, it is assumed that the stream of data is done."""
 
             try:
-                p_id[0], coordinates, obj = stream_iter.next()
+                p_id[0], coordinates, obj = next(stream_iter)
             except StopIteration:
                # we're done
                return -1
@@ -687,8 +698,8 @@ class Index(object):
             return 0
 
 
-        next = core.NEXTFUNC(py_next_item)
-        return core.rt.Index_CreateWithStream(self.properties.handle, next)
+        stream = core.NEXTFUNC(py_next_item)
+        return core.rt.Index_CreateWithStream(self.properties.handle, stream)
 
     def leaves(self):
         leaf_node_count = ctypes.c_uint32()
@@ -793,12 +804,27 @@ class Property(object):
     settable index properties.  Many of these properties must be set at
     index creation times, while others can be used to adjust performance
     or behavior."""
-    def __init__(self, handle = None, owned=True):
+    
+    pkeys = (
+        'buffering_capacity', 'custom_storage_callbacks', 
+        'custom_storage_callbacks_size', 'dat_extension', 'dimension', 
+        'filename', 'fill_factor', 'idx_extension', 'index_capacity', 
+        'index_id', 'leaf_capacity', 'near_minimum_overlap_factor', 
+        'overwrite', 'pagesize', 'point_pool_capacity', 
+        'region_pool_capacity', 'reinsert_factor', 
+        'split_distribution_factor', 'storage', 'tight_mbr', 'tpr_horizon',
+        'type', 'variant', 'writethrough' )
+
+    def __init__(self, handle=None, owned=True, **kwargs):
         if handle:
             self.handle = handle
         else:
             self.handle = core.rt.IndexProperty_Create()
         self.owned = owned
+        for k, v in list(kwargs.items()):
+            if v is not None:
+                setattr(self, k, v)
+
     def __del__(self):
         if self.owned:
             if self.handle and core:
@@ -809,6 +835,23 @@ class Property(object):
                     # but for some reason the dll isn't active
                     return
                 core.rt.IndexProperty_Destroy(self.handle)
+
+    def as_dict(self):
+        d = {}
+        for k in self.pkeys:
+            try:
+                v = getattr(self, k)
+            except core.RTreeError:
+                v = None
+            d[k] = v
+        return d
+
+    def __repr__(self):
+        return repr(self.as_dict())
+
+    def __str__(self):
+        return pprint.pformat(self.as_dict())
+        
     def get_index_type(self):
         return core.rt.IndexProperty_GetIndexType(self.handle)
     def set_index_type(self, value):
@@ -1051,6 +1094,10 @@ class CustomStorageCallbacks(ctypes.Structure):
     destroyCallbackType = ctypes.CFUNCTYPE(
                             None, ctypes.c_void_p, ctypes.POINTER(ctypes.c_int)
                           )
+    flushCallbackType = ctypes.CFUNCTYPE(
+                          None, ctypes.c_void_p, ctypes.POINTER(ctypes.c_int)
+                        )
+
     loadCallbackType    = ctypes.CFUNCTYPE(
                             None, ctypes.c_void_p, id_type, ctypes.POINTER(ctypes.c_uint32),
                             ctypes.POINTER(ctypes.POINTER(ctypes.c_uint8)), ctypes.POINTER(ctypes.c_int)
@@ -1066,16 +1113,18 @@ class CustomStorageCallbacks(ctypes.Structure):
     _fields_ = [ ('context', ctypes.c_void_p),
                  ('createCallback', createCallbackType),
                  ('destroyCallback', destroyCallbackType),
+                 ('flushCallback', flushCallbackType),
                  ('loadCallback', loadCallbackType),
                  ('storeCallback', storeCallbackType),
                  ('deleteCallback', deleteCallbackType),
                ]
 
-    def __init__(self, context, createCallback, destroyCallback, loadCallback, storeCallback, deleteCallback):
+    def __init__(self, context, createCallback, destroyCallback, flushCallback, loadCallback, storeCallback, deleteCallback):
         ctypes.Structure.__init__( self,
                                    ctypes.c_void_p( context ),
                                    self.createCallbackType( createCallback ),
-                                   self.createCallbackType( destroyCallback ),
+                                   self.destroyCallbackType( destroyCallback ),
+                                   self.flushCallbackType ( flushCallback ),
                                    self.loadCallbackType  ( loadCallback ),
                                    self.storeCallbackType ( storeCallback ),
                                    self.deleteCallbackType( deleteCallback ),
@@ -1110,8 +1159,10 @@ class CustomStorageBase(ICustomStorage):
     """
 
     def registerCallbacks(self, properties):
-        callbacks = CustomStorageCallbacks( 0, self.create, self.destroy, self.loadByteArray,
-                                               self.storeByteArray, self.deleteByteArray )
+        callbacks = CustomStorageCallbacks( ctypes.c_void_p(), self.create, 
+                                            self.destroy, self.flush,
+                                            self.loadByteArray, self.storeByteArray, 
+                                            self.deleteByteArray )
         properties.custom_storage_callbacks_size = ctypes.sizeof( callbacks )
         self.callbacks = callbacks
         properties.custom_storage_callbacks      = ctypes.cast( ctypes.pointer(callbacks), ctypes.c_void_p )
@@ -1141,6 +1192,10 @@ class CustomStorageBase(ICustomStorage):
         returnError.contents.value = self.IllegalStateError
         raise NotImplementedError( "You must override this method." )
 
+    def flush(self, context, returnError):
+        """ please override """
+        returnError.contents.value = self.IllegalStateError
+        raise NotImplementedError( "You must override this method." )
 
 
 class CustomStorage(ICustomStorage):
@@ -1151,7 +1206,7 @@ class CustomStorage(ICustomStorage):
     """
 
     def registerCallbacks(self, properties):
-        callbacks = CustomStorageCallbacks( 0, self._create, self._destroy, self._loadByteArray,
+        callbacks = CustomStorageCallbacks( 0, self._create, self._destroy, self._flush, self._loadByteArray,
                                                self._storeByteArray, self._deleteByteArray )
         properties.custom_storage_callbacks_size = ctypes.sizeof( callbacks )
         self.callbacks = callbacks
@@ -1165,6 +1220,9 @@ class CustomStorage(ICustomStorage):
     def _destroy(self, context, returnError):
         self.destroy( returnError )
 
+    def _flush(self, context, returnError):
+        self.flush( returnError )
+
     def _loadByteArray(self, context, page, resultLen, resultData, returnError):
         resultString = self.loadByteArray( page, returnError )
         if returnError.contents.value != self.NoError:
@@ -1201,6 +1259,11 @@ class CustomStorage(ICustomStorage):
         returnError.contents.value = self.IllegalStateError
         raise NotImplementedError( "You must override this method." )
 
+    def flush(self, returnError):
+        """ Must be overriden. No return value. """
+        returnError.contents.value = self.IllegalStateError
+        raise NotImplementedError( "You must override this method." )
+
     def loadByteArray(self, page, returnError):
         """ Must be overriden. Must return a string with the loaded data. """
         returnError.contents.value = self.IllegalStateError
